
#include "parser.h"
#include "preprocess.h"
#include "exec.h"
#include "hypergraph.h"
#include "cost_table.h"
#include "enumerators.h"

#include <sstream>
#include <fstream>

#define PCC_TPTP_SYNTAX
// #define DO_TEST_PCC

BEGIN_HSPS_NAMESPACE

struct ppc {
  const Name* name;
  NTYPE     weight;
  plan_constraint_type pct;
  index_set s_c;
  index_set s_t;
  PDDL_Base::Preference* src;

  ppc()
    : name(0), weight(0), pct(pc_at_end), src(0) { };
  ppc(Name* n, NTYPE w, plan_constraint_type t, const index_set& a_c, const index_set& a_t, PDDL_Base::Preference* s)
    : name(n), weight(w), pct(t), s_c(a_c), s_t(a_t), src(s) { };

  ppc& operator=(const ppc& p) {
    name = p.name;
    weight = p.weight;
    pct = p.pct;
    s_c.assign_copy(p.s_c);
    s_t.assign_copy(p.s_t);
    src = p.src;
  };
};

void write_ppc(std::ostream& s, const Instance& ins, const ppc& p)
{
  if (p.name)
    s << p.name << ": ";
  switch (p.pct) {
  case pc_at_end:
    s << "at-end ";
    ins.write_atom_set(s, p.s_c);
    break;
  case pc_always:
    s << "always ";
    ins.write_atom_set(s, p.s_c);
    break;
  case pc_sometime:
    s << "sometime ";
    ins.write_atom_set(s, p.s_c);
    break;
  case pc_at_most_once:
    s << "at-most-once ";
    ins.write_atom_set(s, p.s_c);
    break;
  case pc_sometime_before:
    s << "sometime-before ";
    ins.write_atom_set(s, p.s_t);
    s << " ";
    ins.write_atom_set(s, p.s_c);
    break;
  case pc_sometime_after:
    s << "sometime-after ";
    ins.write_atom_set(s, p.s_t);
    s << " ";
    ins.write_atom_set(s, p.s_c);
    break;
  default:
    s << "ERROR!";
    exit(255);
  }
  s << " (" << p.weight << ")";
}

void write_ppc_as_pcc(std::ostream& s, const Instance& ins, const ppc& p)
{
  switch (p.pct) {
#ifdef PCC_TPTP_SYNTAX
  case pc_at_end:
    assert(p.s_c.length() == 1);
    s << "fof(goal" << p.s_c[0] << ",axiom,at_end(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
  case pc_always:
    assert(p.s_c.length() == 1);
    s << "fof(ac" << p.s_c[0] << ",axiom,always(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
  case pc_sometime:
    assert(p.s_c.length() == 1);
    s << "fof(ec" << p.s_c[0] << ",axiom,sometime(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
  case pc_at_most_once:
    assert(p.s_c.length() == 1);
    s << "fof(amoc" << p.s_c[0] << ",axiom,at_most_once(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
  case pc_sometime_before:
    assert(p.s_t.length() == 1);
    assert(p.s_c.length() == 1);
    s << "fof(sbc" << p.s_t[0] << "_" << p.s_c[0]
      << ",axiom,sometime_before(";
    ins.atoms[p.s_t[0]].name->write(s, Name::NC_INSTANCE);
    s << ",";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
  case pc_sometime_after:
    assert(p.s_t.length() == 1);
    assert(p.s_c.length() == 1);
    s << "fof(sac" << p.s_t[0] << "_" << p.s_c[0]
      << ",axiom,sometime_after(";
    ins.atoms[p.s_t[0]].name->write(s, Name::NC_INSTANCE);
    s << ",";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
    break;
#else
  case pc_at_end:
    assert(p.s_c.length() == 1);
    s << "AtEnd(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
  case pc_always:
    assert(p.s_c.length() == 1);
    s << "Always(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
  case pc_sometime:
    assert(p.s_c.length() == 1);
    s << "Sometime(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
  case pc_at_most_once:
    assert(p.s_c.length() == 1);
    s << "AtMostOnce(";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
  case pc_sometime_before:
    assert(p.s_t.length() == 1);
    assert(p.s_c.length() == 1);
    s << "SometimeBefore(";
    ins.atoms[p.s_t[0]].name->write(s, Name::NC_INSTANCE);
    s << ",";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
  case pc_sometime_after:
    assert(p.s_t.length() == 1);
    assert(p.s_c.length() == 1);
    s << "SometimeAfter(";
    ins.atoms[p.s_t[0]].name->write(s, Name::NC_INSTANCE);
    s << ",";
    ins.atoms[p.s_c[0]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
    break;
#endif
  default:
    s << "ERROR!";
    exit(255);
  }
}

void enforce_ppc(const ppc& p, Instance& ins)
{
  switch (p.pct) {
  case pc_always:
    ins.enforce_pc_always_conjunction(p.s_c, p.name);
    break;
  case pc_sometime:
    ins.enforce_pc_sometime_conjunction(p.s_c, p.name);
    break;
  case pc_at_most_once:
    ins.enforce_pc_at_most_once_conjunction(p.s_c, p.name);
    break;
  case pc_sometime_before:
    ins.enforce_pc_sometime_before_cc(p.s_t, p.s_c, p.name);
    break;
  default:
    std::cerr << "ERROR!";
    exit(255);
  }
}

bool test_ppc(const ppc& p, ExecTrace* t)
{
  switch (p.pct) {
  case pc_always:
    return t->test_always_conjunction(p.s_c);
  case pc_sometime:
    return t->test_sometime_conjunction(p.s_c);
  case pc_at_most_once:
    return t->test_at_most_once_conjunction(p.s_c);
  case pc_sometime_before:
    if (p.s_c.length() != 1) return false;
    if (p.s_t.length() != 1) return false;
    return t->test_sometime_before(p.s_t[0], p.s_c[0]);
  default:
    std::cerr << "ERROR!";
    exit(255);
  }
}

typedef lvector<ppc> ppc_vec;

void write_ppc_set(std::ostream& s, const ppc_vec& pv, const index_set& set)
{
  s << "{";
  for (index_type k = 0; k < set.length(); k++) {
    if (k > 0) s << ", ";
    s << pv[set[k]].name;
  }
  s << "}";
}

void write_ppc_set_as_pcc
(std::ostream& s, Instance& ins, const ppc_vec& pv, const index_set& set,
 bool has_val, NTYPE set_val, NTYPE set_pen)
{
  s << "%% set {";
  for (index_type i = 0; i < set.length(); i++) {
    if (i > 0) s << ", ";
    s << pv[set[i]].name;
  }
  if (has_val)
    s << "}: value = " << PRINT_NTYPE(set_val)
      << ", penalty = " << PRINT_NTYPE(set_pen)
      << std::endl;
  else
    s << "}" << std::endl;
  for (index_type i = 0; i < set.length(); i++)
    write_ppc_as_pcc(s, ins, pv[set[i]]);
  for (index_type i = 0; i < ins.goal_atoms.length(); i++) {
#ifdef PCC_TPTP_SYNTAX
    s << "fof(goal" << ins.goal_atoms[i] << ",axiom,at_end(";
    ins.atoms[ins.goal_atoms[i]].name->write(s, Name::NC_INSTANCE);
    s << "))." << std::endl;
#else
    s << "AtEnd(";
    ins.atoms[ins.goal_atoms[i]].name->write(s, Name::NC_INSTANCE);
    s << ")." << std::endl;
#endif
  }
}

bool append_ppc
(PDDL_Base* b,
 PDDL_Base::Preference* p,
 PDDL_Base::Goal* g1,
 PDDL_Base::Goal* g2,
 Instance& ins,
 plan_constraint_type t,
 index_type k,
 ppc_vec& ppcv)
{
  PDDL_Base::signed_atom_vec a;
  bool g1_is_dis;
  index_set s1;
  if (b->goal_to_atom_vec(g1, a, g1_is_dis)) if (!g1_is_dis) {
    Name* n = (p->name ?
	       (Name*)new StringName(p->name->print_name) :
	       (Name*)new EnumName("ppc", k));
    NTYPE w = p->value(b->metric_type, b->metric);
    b->instantiate_atom_set(ins, a, s1);
    for (index_type i = 0; i < s1.length(); i++)
      ins.atoms[s1[i]].goal = true;
    if (g2) {
      PDDL_Base::signed_atom_vec a2;
      bool g2_is_dis;
      index_set s2;
      if (b->goal_to_atom_vec(g2, a2, g2_is_dis)) if (!g2_is_dis) {
	b->instantiate_atom_set(ins, a2, s2);
	for (index_type i = 0; i < s2.length(); i++)
	  ins.atoms[s2[i]].goal = true;
	ppcv.append(ppc(n, w, t, s1, s2, p));
      }
      else {
	std::cerr << "error: can't handle goal ";
	p->print(std::cerr);
	std::cerr << std::endl;
	return false;
      }
    }
    else {
      ppcv.append(ppc(n, w, t, s1, EMPTYSET, p));
    }
    return true;
  }
  else {
    std::cerr << "error: can't handle goal ";
    p->print(std::cerr);
    std::cerr << std::endl;
    return false;
  }
}

class TestPCC {
  Instance& ins;
  Preprocessor prep;
  bool print_inf;
  bool opt_H2;

  // input (and some derived) constraints
  index_set c_always;
  index_set c_never;
  index_set c_sometime;
  index_set c_at_most_once;
  pair_set  c_sometime_before;

  bool_vec  allowed_acts;
  bool      aa_changed;
  bool_vec  disallowed_atms;
  bool_vec  reachable_atms;
  bool_vec  reachable_acts;

  // note: sb holds the inverse of the sometime-before relation,
  // i.e. p -> q iff sometime-before(q, p)
  graph sb;
  bool_vec as;
  graph asb;

  void assert_always(index_type i);
  void assert_never(index_type i);
  void update_reachability();
  void update_precedence();
  void infer_never_from_sb();
  void assert_sometime(index_type i);
  bool check_sometime_never();

public:
  TestPCC(Instance& i, Statistics& st,
	  ppc_vec& v, const index_set& s,
	  bool optH2, bool opt_print_inf);
  bool run();
};

TestPCC::TestPCC
(Instance& i, Statistics& st, ppc_vec& v, const index_set& s,
 bool optH2, bool opt_print_inf)
  : ins(i), prep(i, st), print_inf(opt_print_inf), opt_H2(optH2)
{
  for (index_type k = 0; k < s.length(); k++) {
    if (v[s[k]].pct == pc_always) {
      assert(v[s[k]].s_c.length() == 1);
      c_always.insert(v[s[k]].s_c);
    }
    else if (v[s[k]].pct == pc_sometime) {
      assert(v[s[k]].s_c.length() == 1);
      c_sometime.insert(v[s[k]].s_c);
    }
    else if (v[s[k]].pct == pc_at_most_once) {
      assert(v[s[k]].s_c.length() == 1);
      c_at_most_once.insert(v[s[k]].s_c);
    }
    else if (v[s[k]].pct == pc_sometime_before) {
      assert(v[s[k]].s_t.length() == 1);
      assert(v[s[k]].s_c.length() == 1);
      c_sometime_before.insert(index_pair(v[s[k]].s_c[0],v[s[k]].s_t[0]));
    }
  }
  allowed_acts.assign_value(true, ins.n_actions());
  disallowed_atms.assign_value(false, ins.n_atoms());
  prep.compute_reachability(reachable_atms, reachable_acts,
			    &allowed_acts, opt_H2);
  prep.quick_landmark_graph(sb, &reachable_atms, &allowed_acts, opt_H2);
  aa_changed = false;
}

void TestPCC::assert_always(index_type i)
{
  for (index_type k = 0; k < ins.atoms[i].del_by.length(); k++)
    if (allowed_acts[ins.atoms[i].del_by[k]]) {
      if (print_inf) {
	std::cout << "Always(" << ins.atoms[i].name << ") => -Allowed("
		  << ins.actions[ins.atoms[i].del_by[k]].name << ")"
		  << std::endl;
      }
      aa_changed = true;
      allowed_acts[ins.atoms[i].del_by[k]] = false;
    }
}

void TestPCC::assert_never(index_type i)
{
  for (index_type k = 0; k < ins.atoms[i].add_by.length(); k++)
    if (allowed_acts[ins.atoms[i].add_by[k]]) {
      if (print_inf) {
	std::cout << "Never(" << ins.atoms[i].name << ") => -Allowed("
		  << ins.actions[ins.atoms[i].add_by[k]].name << ")"
		  << std::endl;
      }
      aa_changed = true;
      allowed_acts[ins.atoms[i].add_by[k]] = false;
    }
  disallowed_atms[i] = true;
}

void TestPCC::update_reachability()
{
  bool_vec ratms;
  bool_vec racts;
  prep.compute_reachability(ratms, racts, &allowed_acts, opt_H2);
  for (index_type k = 0; k < ins.n_atoms(); k++)
    if (reachable_atms[k] && !ratms[k]) {
      if (print_inf)
	std::cout << "-Reachable(" << ins.atoms[k].name << ")" << std::endl;
      reachable_atms[k] = false;
    }
  for (index_type k = 0; k < ins.n_atoms(); k++)
    if (reachable_acts[k] && !racts[k]) {
      if (print_inf)
	std::cout << "-Reachable(" << ins.actions[k].name << ")" << std::endl;
      reachable_acts[k] = false;
    }
  aa_changed = false;
}

void TestPCC::update_precedence()
{
  graph sb1;
  prep.quick_landmark_graph(sb1, &reachable_atms, &allowed_acts, opt_H2);
  for (index_type i = 0; i < ins.n_atoms(); i++)
    for (index_type j = 0; j < ins.n_atoms(); j++) {
      if (sb1.adjacent(i, j) && !sb.adjacent(i, j)) {
	if (print_inf) {
	  std::cout << "Landmark(" << ins.atoms[i].name
		    << "," << ins.atoms[j].name << ") => SometimeBefore("
		    << ins.atoms[i].name << "," << ins.atoms[j].name
		    << ")" << std::endl;
	}
	sb.add_edge(i, j);
      }
    }
}

void TestPCC::infer_never_from_sb()
{
  sb.transitive_closure();
  for (index_type i = 0; i < ins.n_atoms(); i++)
    for (index_type j = i + 1; j < ins.n_atoms(); j++)
      if (sb.adjacent(i, j) && sb.adjacent(j, i)) {
	if (!disallowed_atms[i] || !disallowed_atms[j]) {
	  if (print_inf) {
	    std::cout << "SometimeBefore(" << ins.atoms[i].name << ","
		      << ins.atoms[j].name << ") & SometimeBefore("
		      << ins.atoms[j].name << "," << ins.atoms[i].name
		      << ") => Never(" << ins.atoms[i].name << ") & Never("
		      << ins.atoms[j].name << ")" << std::endl;
	  }
	  assert_never(i);
	  assert_never(j);
	}
      }
}

void TestPCC::assert_sometime(index_type i)
{
  if (!as[i]) {
    as[i] = true;
    index_set b;
    sb.ancestors(i, b);
    asb.add_edge(b, i);
    for (index_type k = 0; k < b.length(); k++) if (!as[b[k]]) {
      if (print_inf) {
	std::cout << "Sometime(" << ins.atoms[i].name
		  << ") & SometimeBefore(" << ins.atoms[i].name
		  << "," << ins.atoms[b[k]].name << ") => Sometime("
		  << ins.atoms[b[k]].name << ")" << std::endl;
      }
      assert_sometime(b[k]);
    }
  }
}

bool TestPCC::check_sometime_never()
{
  for (index_type k = 0; k < ins.n_atoms(); k++) if (as[k]) {
    if (disallowed_atms[k]) {
      if (print_inf) {
	std::cout << "Sometime(" << ins.atoms[k].name << ") & Never("
		  << ins.atoms[k].name << ") => CONTRADICTION"
		  << std::endl;
      }
      return true;
    }
    else if (!reachable_atms[k]) {
      if (print_inf) {
	std::cout << "Sometime(" << ins.atoms[k].name
		  << ") & -Reachable(" << ins.atoms[k].name
		  << ") => CONTRADICTION" << std::endl;
      }
      return true;
    }
  }
  return false;
}

bool TestPCC::run()
{
  std::cerr << "asserting given always constraints..." << std::endl;
  for (index_type k = 0; k < c_always.length(); k++)
    assert_always(c_always[k]);
  if (aa_changed) update_reachability();
  std::cerr << "asserting given sometime-before constraints..." << std::endl;
  for (index_type k = 0; k < c_sometime_before.length(); k++)
    sb.add_edge(c_sometime_before[k].first, c_sometime_before[k].second);

  std::cerr << "infering never from cyclic sb's..." << std::endl;
  infer_never_from_sb();
  while (aa_changed) {
    update_reachability();
    update_precedence();
    infer_never_from_sb();
  }

  as.assign_value(false, ins.n_atoms());
  asb.init(ins.n_atoms());

  std::cerr << "asserting goals..." << std::endl;
  for (index_type k = 0; k < ins.goal_atoms.length(); k++)
    assert_sometime(ins.goal_atoms[k]);
  std::cerr << "asserting given sometime constraints..." << std::endl;
  for (index_type k = 0; k < c_sometime.length(); k++)
    assert_sometime(c_sometime[k]);

  bool con = check_sometime_never();
  if (con) return true;

  std::cerr << "failed to prove contradiction" << std::endl;
  return false;
}

int main(int argc, char *argv[]) {
  bool     opt_preprocess = true;
  bool     opt_test = true;
  bool     opt_max = true;
  bool     opt_min_inc = false;
  bool     opt_max_sat = false;
  bool     opt_H2 = true;
  bool     opt_H3 = false;
  bool     opt_split = false;
  bool     opt_precheck = false;
  bool     opt_load = true;
  bool     opt_pddl = true;
  bool     opt_untimed = false;
  bool     opt_save_test = false;
  bool     opt_save_max = false;
  bool     opt_save_open = false;
  bool     opt_save_ext = false;
  bool     opt_save_compiled = false;
  NTYPE    opt_slack = 0;
  bool     opt_pcc = false;
  bool     opt_verbose = false;
  index_type opt_min_d = 1;
  index_type opt_max_d = 2;
  unsigned int random_seed = 0;

  count_type file_write_limit = 1000;
  count_type files_written = 0;

  PDDL_Base::write_warnings = false;
  PDDL_Base::compile_away_plan_constraints = false;

  Instance::write_PDDL2 = false;
  Instance::write_PDDL3 = false;
  Instance::write_metric = false;
  Instance::write_time = false;

  Statistics stats;
  stats.enable_interrupt();
  stats.start();

  StringTable symbols(50, lowercase_map);
  Parser* reader = new Parser(symbols);

  for (int k = 1; k < argc; k++) {
    if ((strcmp(argv[k],"-v") == 0) && (k < argc - 1)) {
      Instance::default_trace_level = atoi(argv[++k]);
      opt_verbose = true;
    }
    else if (strcmp(argv[k],"-no-prep") == 0) {
      opt_preprocess = false;
    }
    else if ((strcmp(argv[k],"-catc") == 0) && (k < argc - 1)) {
      HSPS::PDDL_Name::catc = *argv[++k];
    }
    else if (strcmp(argv[k],"-3") == 0) {
      opt_H3 = true;
    }
    else if (strcmp(argv[k],"-1") == 0) {
      opt_H2 = false;
    }
    else if ((strcmp(argv[k],"-d-min") == 0) && (k < argc - 1)) {
      opt_min_d = atoi(argv[++k]);
    }
    else if ((strcmp(argv[k],"-d-max") == 0) && (k < argc - 1)) {
      opt_max_d = atoi(argv[++k]);
    }
    else if (strcmp(argv[k],"-no-test") == 0) {
      opt_test = false;
    }
    else if (strcmp(argv[k],"-no-max") == 0) {
      opt_max = false;
    }
    else if (strcmp(argv[k],"-no-pddl") == 0) {
      opt_pddl = false;
    }
    else if (strcmp(argv[k],"-load") == 0) {
      opt_load = true;
    }
    else if (strcmp(argv[k],"-no-load") == 0) {
      opt_load = false;
    }
    else if (strcmp(argv[k],"-untimed") == 0) {
      opt_untimed = true;
    }
    else if (strcmp(argv[k],"-split") == 0) {
      opt_split = true;
    }
    else if (strcmp(argv[k],"-precheck") == 0) {
      opt_precheck = true;
    }
    else if ((strcmp(argv[k],"-save-test") == 0) ||
	     (strcmp(argv[k],"-save-tests") == 0)) {
      opt_save_test = true;
    }
    else if ((strcmp(argv[k],"-save-ext") == 0) ||
	     (strcmp(argv[k],"-save-exts") == 0)) {
      opt_save_ext = true;
    }
    else if (strcmp(argv[k],"-save-max") == 0) {
      opt_save_max = true;
    }
    else if (strcmp(argv[k],"-save-open") == 0) {
      opt_save_open = true;
    }
    else if (strcmp(argv[k],"-save-compiled") == 0) {
      opt_save_compiled = true;
    }
    else if (strcmp(argv[k],"-pcc") == 0) {
      opt_pcc = true;
    }
    else if (strcmp(argv[k],"-min-inc-pcc") == 0) {
      opt_pcc = true;
      opt_min_inc = true;
    }
    else if (strcmp(argv[k],"-max-sat-pcc") == 0) {
      opt_pcc = true;
      opt_max_sat = true;
    }
    else if ((strcmp(argv[k],"-slack") == 0) && (k < argc - 1)) {
      opt_slack = A_TO_N(argv[++k]);
    }
    else if (strcmp(argv[k],"-max-slack") == 0) {
      opt_slack = POS_INF;
    }
    else if (strcmp(argv[k],"-no-dkel") == 0) {
      HSPS::Instance::write_DKEL = false;
    }
    else if (strcmp(argv[k],"-no-negation") == 0) {
      HSPS::Instance::write_negation = false;
    }
    else if ((strcmp(argv[k],"-file-limit") == 0) && (k < argc - 1)) {
      file_write_limit = atoi(argv[++k]);
    }
    else if (((strcmp(argv[k],"-rnd") == 0) ||
	      (strcmp(argv[k],"-r") == 0)) &&
	     (k < argc - 1)) {
      random_seed = atoi(argv[++k]);
    }
    else if (*argv[k] != '-') {
      reader->read(argv[k], false);
    }
  }

  Instance instance;
  ppc_vec  ppcs;

  reader->post_process();

  stats.start();
  reader->instantiate(instance);
  index_set hard_goals;
  for (index_type k = 0; k < instance.n_atoms(); k++)
    if (instance.atoms[k].goal)
      hard_goals.insert(k);

  bool ok = true;
  for (index_type k = 0; k < reader->dom_preferences.length(); k++) {
    if (reader->dom_preferences[k]->goal->g_class == PDDL_Base::goal_always) {
      if (!append_ppc(reader,
		      reader->dom_preferences[k],
		      ((PDDL_Base::SimpleSequenceGoal*)reader->dom_preferences[k]->goal)->constraint,
		      0,
		      instance,
		      pc_always,
		      k,
		      ppcs))
	ok = false;
    }
    else if (reader->dom_preferences[k]->goal->g_class == PDDL_Base::goal_sometime) {
      if (!append_ppc(reader,
		      reader->dom_preferences[k],
		      ((PDDL_Base::SimpleSequenceGoal*)reader->dom_preferences[k]->goal)->constraint,
		      0,
		      instance,
		      pc_sometime,
		      k,
		      ppcs))
	ok = false;
    }
    else if (reader->dom_preferences[k]->goal->g_class == PDDL_Base::goal_at_most_once) {
      if (!append_ppc(reader,
		      reader->dom_preferences[k],
		      ((PDDL_Base::SimpleSequenceGoal*)reader->dom_preferences[k]->goal)->constraint,
		      0,
		      instance,
		      pc_at_most_once,
		      k,
		      ppcs))
	ok = false;
    }
    else if (reader->dom_preferences[k]->goal->g_class == PDDL_Base::goal_sometime_before) {
      if (!append_ppc(reader,
		      reader->dom_preferences[k],
		      ((PDDL_Base::TriggeredSequenceGoal*)reader->dom_preferences[k]->goal)->constraint,
		      ((PDDL_Base::TriggeredSequenceGoal*)reader->dom_preferences[k]->goal)->trigger,
		      instance,
		      pc_sometime_before,
		      k,
		      ppcs))
	ok = false;
    }
    else {
      if (!append_ppc(reader,
		      reader->dom_preferences[k],
		      reader->dom_preferences[k]->goal,
		      0,
		      instance,
		      pc_always,
		      k,
		      ppcs))
	ok = false;
    }
  }

  if (!ok) exit(255);

  Preprocessor prep(instance, stats);

  if (opt_preprocess) {
    prep.preprocess();
    for (index_type k = 0; k < ppcs.length(); k++) {
      instance.remap_set(ppcs[k].s_c, prep.atom_map);
      instance.remap_set(ppcs[k].s_t, prep.atom_map);
    }
    instance.remap_set(hard_goals, prep.atom_map);
  }
  else {
    instance.cross_reference();
  }
  instance.set_goal(hard_goals);

  stats.stop();
  std::cerr << "instantiation and preprocessing finished in "
	    << stats.time() << " seconds" << std::endl;

  NTYPE v_sum = 0;
  for (index_type k = 0; k < ppcs.length(); k++)
    v_sum += ppcs[k].weight;
  NTYPE v_max = v_sum;
  NTYPE p_min = 0;

  std::cerr << ppcs.length() << " preferences over plan constraints"
	    << std::endl;
  std::cerr << "sum penalty = " << v_sum << std::endl;
  std::cerr << instance.n_atoms() << " atoms and "
	    << instance.n_actions() << " actions"
	    << std::endl;

  if (opt_verbose)
    for (index_type k = 0; k < ppcs.length(); k++) {
      std::cerr << k << ". ";
      write_ppc(std::cerr, instance, ppcs[k]);
      std::cerr << std::endl;
    }

  // hypergraph over ppcs
  hypergraph h_inc(ppcs.length());
  index_set_vec min_inc;

  if (opt_load) {
    PDDL_Base::write_warnings = false;
    name_vec ppc_name(0, ppcs.length());
    for (index_type k = 0; k < ppcs.length(); k++)
      ppc_name[k] = ppcs[k].name;
    index_set_vec h_load;
    reader->export_sets(ppc_name, h_load);
    for (index_type k = 0; k < h_load.length(); k++)
      min_inc.insert_minimal(h_load[k]);
    std::cerr << min_inc.length()
	      << " minimal inconsistent sets found in input"
	      << std::endl;
  }

  if (opt_min_inc && opt_pcc) {
    for (index_type k = 0; k < min_inc.length(); k++) {
      write_ppc_set_as_pcc(std::cout, instance, ppcs, min_inc[k], false, 0, 0);
#ifdef DO_TEST_PCC
      TestPCC test(instance, stats, ppcs, min_inc[k], opt_H2, opt_verbose);
      bool cp = test.run();
      if (cp)
	std::cout << "contradiction proven" << std::endl;
      else
	std::cout << "contradiction not proven" << std::endl;
#endif
    }
    exit(0);
  }

  for (index_type k = 0; k < min_inc.length(); k++)
    if (h_inc.is_independent(min_inc[k]))
      h_inc.add_edge(min_inc[k]);

  // maximal known-to-be-satisfiable sets of ppcs
  index_set_vec max_sat;

  if (reader->n_plans() > 0) {
    std::cerr << "analyzing " << reader->n_plans() << " plans..." << std::endl;
    NTYPE v_max_plan = 0;

    for (index_type k = 0; k < reader->n_plans(); k++) {
      Schedule* plan = new Schedule(instance);
      reader->export_plan(k, instance, prep.action_map, *plan);
      ExecTrace* trace = new ExecTrace(instance);
      ExecErrorSet* errors = new ExecErrorSet();
      bool ok = plan->simulate(trace, errors, false);
      if (!ok) {
	std::cerr << "plan " << k << " not executable: ";
	errors->write(std::cerr);
	std::cerr << std::endl;
      }
      else {
	if (opt_untimed) {
	  ExecTrace* t1 = trace;
	  trace = t1->stable_trace();
	  delete t1;
	}
	index_set sat;
	NTYPE val = 0;
	for (index_type i = 0; i < ppcs.length(); i++)
	  if (test_ppc(ppcs[i], trace)) {
	    std::cerr << "plan #" << k << " satisfies #" << i << ": ";
	    write_ppc(std::cerr, instance, ppcs[i]);
	    std::cerr << std::endl;
	    sat.insert(i);
	    val += ppcs[i].weight;
	  }
	if (val > v_max_plan) v_max_plan = val;
	max_sat.insert_maximal(sat);
      }
      delete errors;
      delete trace;
      delete plan;
    }

    std::cerr << "maximal satisfied sets of constraints:" << std::endl;
    for (index_type k = 0; k < max_sat.length(); k++) {
      std::cerr << k + 1 << ": ";
      write_ppc_set(std::cerr, ppcs, max_sat[k]);
      std::cerr << std::endl;
    }
    std::cerr << "max plan value = " << PRINT_NTYPE(v_max_plan)
	      << std::endl
	      << "min plan penalty = " << PRINT_NTYPE(v_sum - v_max_plan)
	      << std::endl;
  }

  if (opt_max_sat && opt_pcc) {
    for (index_type k = 0; k < max_sat.length(); k++) {
      write_ppc_set_as_pcc(std::cout, instance, ppcs, max_sat[k], false, 0, 0);
    }
    exit(0);
  }

  // test-related stuff

  index_set ce;
  index_set ca;

  count_type n_tests = 0;

  if (opt_split) {
    for (index_type k = 0; k < ppcs.length(); k++) {
      if (ppcs[k].pct == pc_sometime)
	ce.insert(k);
      else
	ca.insert(k);
    }
    std::cerr << ce.length() << " type-E constraints: " << ce << std::endl;
    std::cerr << ca.length() << " type-A constraints: " << ca << std::endl;
  }
  else {
    ca.fill(ppcs.length());
  }

  if (opt_test) {
    stats.start();

    if (opt_precheck) {
      std::cerr << "pre-checking..." << std::endl;
      CostTable* h0 = new CostTable(instance, stats);
      if (opt_H3) {
	h0->compute_H3(ZeroACF());
      }
      else if (opt_H2) {
	h0->compute_H2(ZeroACF());
      }
      else {
	h0->compute_H1(ZeroACF());
      }
      for (index_type i = 0; i < ppcs.length(); i++)
	if (ppcs[i].pct == pc_always)
	  for (index_type j = 0; j < ppcs.length(); j++)
	    if (ppcs[j].pct == pc_sometime) {
	      index_set s(ppcs[i].s_c);
	      s.insert(ppcs[j].s_c);
	      if (INFINITE(h0->eval(s))) {
		index_set sel;
		sel.insert(i);
		sel.insert(j);
		std::cerr << "set " << sel << " = ";
		write_ppc_set(std::cerr, ppcs, sel);
		std::cerr << " is unachievable" << std::endl;
		h_inc.add_edge(sel);
	      }
	    }
      delete h0;
      n_tests += 1;
    }

    if (ce.length() > 0) {
      std::cerr << "checking type-E constraints..." << std::endl;
      Instance* test_ins = instance.copy();
      index_vec ace(no_such_index, ce.length());
      for (index_type k = 0; k < ce.length(); k++) {
	ace[k] =
	  test_ins->compile_pc_sometime_conjunction(ppcs[ce[k]].s_c, ppcs[ce[k]].name);
	test_ins->atoms[ace[k]].goal = false;
      }
      test_ins->cross_reference();
      CostTable* h0 = new CostTable(*test_ins, stats);
      if (opt_H3) {
	h0->compute_H3(ZeroACF());
      }
      else if (opt_H2) {
	h0->compute_H2(ZeroACF());
      }
      else {
	h0->compute_H1(ZeroACF());
      }
      for (index_type k = 0; k < ce.length(); k++) {
	if (opt_verbose) {
	  std::cerr << "checking " << ce[k] << "..." << std::endl;
	}
	if (INFINITE(h0->incremental_eval(test_ins->goal_atoms, ace[k]))) {
	  std::cerr << "constraint " << ppcs[ce[k]].name
		    << " is unachievable" << std::endl;
	  h_inc.add_singleton_edge(ce[k]);
	}
	else if ((opt_H2 || opt_H3) && (opt_max_d > 1)) {
	  index_set g1(test_ins->goal_atoms);
	  g1.insert(ace[k]);
	  for (index_type i = k + 1; i < ce.length(); i++) {
	    if (opt_verbose) {
	      std::cerr << "checking " << ce[k] << " & " << ce[i] << "..."
			<< std::endl;
	    }
	    if (INFINITE(h0->incremental_eval(g1, ace[i]))) {
	      std::cerr << "constraints " << ppcs[ce[k]].name
			<< " and " << ppcs[ce[i]].name
			<< " are jointly unachievable" << std::endl;
	      h_inc.add_binary_edge(ce[k], ce[i]);
	    }
	  }
	}
      }
      delete h0;
      delete test_ins;
      n_tests += 1;
    }

    index_type d = opt_min_d;
    while ((d <= opt_max_d) && (d < ca.length())) {
      std::cerr << "checking " << d << "-consistency..." << std::endl;
      mSubsetEnumerator e(ca.length(), d);
      bool more = e.first();
      while (more) {
	assert(e.current_set_size() == d);
	index_set sel;
	e.current_set(ca, sel);
	if (h_inc.is_independent(sel)) {
	  if (opt_verbose)
	    std::cerr << "testing " << sel << " ("
		      << h_inc.n_edges() << " inc. sets, "
		      << n_tests << " tests, "
		      << stats.time() << " sec., "
		      << stats.peak_memory()
		      << "k)..." << std::endl;
	  Instance* test_ins = instance.copy();
	  for (index_type i = 0; i < sel.length(); i++)
	    enforce_ppc(ppcs[sel[i]], *test_ins);
	  index_vec ace;
	  if (ce.length() > 0) {
	    ace.assign_value(no_such_index, ce.length());
	    for (index_type k = 0; k < ce.length(); k++) {
	      ace[k] = test_ins->compile_pc_sometime_conjunction(ppcs[ce[k]].s_c, ppcs[ce[k]].name);
	      test_ins->atoms[ace[k]].goal = false;
	    }
	  }
	  // Preprocessor* test_prep = new Preprocessor(*test_ins, stats);
	  // test_prep->preprocess();
	  test_ins->cross_reference();
	  CostTable* h0 = new CostTable(*test_ins, stats);
	  if (opt_H3) {
	    h0->compute_H3(ZeroACF());
	  }
	  else if (opt_H2) {
	    h0->compute_H2(ZeroACF());
	  }
	  else {
	    h0->compute_H1(ZeroACF());
	  }
	  if (INFINITE(h0->eval(test_ins->goal_atoms))) {
	    std::cerr << "set " << sel << " = ";
	    write_ppc_set(std::cerr, ppcs, sel);
	    std::cerr << " is unachievable" << std::endl;
	    h_inc.add_edge(sel);
	  }
	  else {
	    if (opt_save_test) {
	      if (files_written++ > file_write_limit) {
		std::cerr << "error: writing too many files!" << std::endl;
		exit(1);
	      }
	      std::ostringstream fname;
	      char* b = reader->problem_file_basename();
	      if (b) fname << b << "-";
	      fname << "test" << n_tests + 1 << "-compiled.pddl";
	      std::ofstream s_out(fname.str().c_str());
	      s_out << ";; " << instance.name << ", test problem #"
		    << (n_tests + 1) << std::endl;
	      write_ppc_set(s_out << ";; selected constraints: ", ppcs, sel);
	      test_ins->write_domain_init(s_out);
	      test_ins->write_domain_declarations(s_out);
	      test_ins->write_domain_actions(s_out);
	      test_ins->write_domain_DKEL_items(s_out);
	      if (HSPS::Instance::write_extra) {
		name_vec set_names(0, 0);
		index_set_vec sets;
		reader->export_action_partitions(set_names, sets);
		s_out << ";; action partitions" << std::endl;
		for (index_type k = 0; k < sets.length(); k++) {
		  instance.remap_set(sets[k], prep.action_map);
		  instance.write_domain_action_set(s_out, sets[k], set_names[k]);
		}
	      }
	      s_out << ")";
	      test_ins->write_problem(s_out);
	    }
	    if (ce.length() > 0) {
	      for (index_type k = 0; k < ce.length(); k++) {
		if (opt_verbose) {
		  std::cerr << "checking " << sel << " & " << ce[k] << "..."
			    << std::endl;
		}
		if (INFINITE(h0->incremental_eval(test_ins->goal_atoms, ace[k]))) {
		  index_set sel1(sel);
		  sel1.insert(ce[k]);
		  if (h_inc.is_independent(sel1)) {
		    std::cerr << "set " << sel1 << " = ";
		    write_ppc_set(std::cerr, ppcs, sel1);
		    std::cerr << " is unachievable" << std::endl;
		    h_inc.add_edge(sel1);
		  }
		}
		else if ((opt_H2 || opt_H3) && (opt_max_d > 1)) {
		  index_set g1(test_ins->goal_atoms);
		  g1.insert(ace[k]);
		  for (index_type i = k + 1; i < ce.length(); i++) {
		    if (opt_verbose) {
		      std::cerr << "checking " << sel << " & " << ce[k]
				<< " & " << ce[i] << "..." << std::endl;
		    }
		    if (INFINITE(h0->incremental_eval(g1, ace[i]))) {
		      index_set sel2(sel);
		      sel2.insert(ce[k]);
		      sel2.insert(ce[i]);
		      if (h_inc.is_independent(sel2)) {
			std::cerr << "set " << sel2 << " = ";
			write_ppc_set(std::cerr, ppcs, sel2);
			std::cerr << " is unachievable" << std::endl;
			h_inc.add_edge(sel2);
		      }
		    }
		  }
		}
	      }
	    }
	  }
	  delete h0;
	  delete test_ins;
	  n_tests += 1;
	}
	more = e.next();
      }
      std::cerr << h_inc.n_edges() << " inconsistent sets found ("
		<< n_tests << " tests, " << stats << ")" << std::endl;
      d += 1;
    }

    stats.stop();
  } // if (opt_test) ...

  if (opt_max) {
    std::cerr << "computing maximal consistent sets..." << std::endl;

    index_set_vec mcs;
    h_inc.independent_sets(mcs);
    cost_vec set_value(0, mcs.length());
    cost_vec set_penalty(0, mcs.length());
    std::cerr << mcs.length() << " maximal consistent sets" << std::endl;
    for (index_type k = 0; k < mcs.length(); k++) {
      if (opt_verbose) std::cerr << "{";
      for (index_type i = 0; i < mcs[k].length(); i++) {
	if (opt_verbose) {
	  if (i > 0) std::cerr << ", ";
	  std::cerr << ppcs[mcs[k][i]].name;
	}
	set_value[k] += ppcs[mcs[k][i]].weight;
      }
      set_penalty[k] = v_sum - set_value[k];
      if (opt_verbose) 
	std::cerr << "}: value = " << PRINT_NTYPE(set_value[k])
		  << ", penalty = " << PRINT_NTYPE(set_penalty[k])
		  << std::endl;
    }

    v_max = cost_vec_util::max(set_value);
    p_min = cost_vec_util::min(set_penalty);

    std::cerr << "max value = " << PRINT_NTYPE(v_max) << std::endl;
    std::cerr << "min penalty = " << PRINT_NTYPE(p_min) << std::endl;

    if (opt_pcc) {
      for (index_type k = 0; k < mcs.length(); k++) {
	if (opt_save_max) {
	  if (set_value[k] >= (v_max - opt_slack)) {
	    if (files_written++ > file_write_limit) {
	      std::cerr << "error: writing too many files!" << std::endl;
	      exit(1);
	    }
	    std::ostringstream fname;
	    char* b = reader->problem_file_basename();
	    if (b) fname << b << "-";
	    fname << "max" << k + 1 << ".in";
	    std::cerr << "writing file " << fname.str() << "..." << std::endl;
	    std::ofstream s_out(fname.str().c_str());
#ifndef PCC_TPTP_SYNTAX
	    s_out << "list(sos)." << std::endl;
#endif
	    write_ppc_set_as_pcc(s_out, instance, ppcs, mcs[k],
				 true, set_value[k], set_penalty[k]);
#ifndef PCC_TPTP_SYNTAX
	    s_out << "end_of_list." << std::endl;
#endif
	    s_out.close();
	  }
	}
	else {
	  write_ppc_set_as_pcc(std::cout, instance, ppcs, mcs[k],
			       true, set_value[k], set_penalty[k]);
	}
      }
    }

    if (opt_save_max && !opt_pcc) {
      for (index_type k = 0; k < mcs.length(); k++)
	if (set_value[k] >= (v_max - opt_slack)) {
	  if (files_written++ > file_write_limit) {
	    std::cerr << "error: writing too many files!" << std::endl;
	    exit(1);
	  }
	  std::ostringstream fname;
	  char* b = reader->problem_file_basename();
	  if (b) fname << b << "-";
	  fname << "max" << k + 1;
	  if (opt_save_compiled) fname << "-compiled";
	  fname << ".pddl";
	  std::cerr << "writing file " << fname.str() << "..." << std::endl;
	  std::ofstream s_out(fname.str().c_str());

	  s_out << ";; " << instance.name << ", test problem #"
		<< (k + 1) << std::endl;
	  s_out << ";; selected constraints: {";
	  for (index_type i = 0; i < mcs[k].length(); i++) {
	    if (i > 0) s_out << ", ";
	    s_out << ppcs[mcs[k][i]].name;
	  }
	  s_out << "}" << std::endl;
	  s_out << ";; value = " << PRINT_NTYPE(set_value[k])
		<< ", penalty = " << PRINT_NTYPE(set_penalty[k])
		<< std::endl;

	  if (opt_save_compiled) {
	    Instance* test_ins = instance.copy();
	    for (index_type i = 0; i < mcs[k].length(); i++)
	      enforce_ppc(ppcs[mcs[k][i]], *test_ins);
	    test_ins->write_domain_init(s_out);
	    test_ins->write_domain_declarations(s_out);
	    test_ins->write_domain_actions(s_out);
	    test_ins->write_domain_DKEL_items(s_out);
	    if (HSPS::Instance::write_extra) {
	      name_vec set_names(0, 0);
	      index_set_vec sets;
	      reader->export_action_partitions(set_names, sets);
	      s_out << ";; action partitions" << std::endl;
	      for (index_type k = 0; k < sets.length(); k++) {
		instance.remap_set(sets[k], prep.action_map);
		instance.write_domain_action_set(s_out, sets[k], set_names[k]);
	      }
	    }
	    s_out << ")";
	    test_ins->write_problem(s_out);
	    delete test_ins;
	  }
	  else {
	    reader->write_problem_begin(s_out);
	    reader->write_objects(s_out, true);
	    reader->write_init(s_out);
	    s_out << " (:goal (and";
	    for (index_type i = 0; i < reader->dom_goals.length(); i++) {
	      if (reader->dom_goals[i]->is_state() &&
		  reader->dom_goals[i]->is_propositional()) {
		s_out << " ";
		reader->dom_goals[i]->print(s_out);
	      }
	    }
	    s_out << "))" << std::endl << " (:constraints (and";
	    for (index_type i = 0; i < mcs[k].length(); i++) {
	      s_out << " ";
	      ppcs[mcs[k][i]].src->goal->print(s_out);
	    }
	    s_out << "))" << std::endl << ")" << std::endl;
	  }
	  s_out.close();
	}
    }
  } // if (opt_max) ...

  if (opt_save_open) {
    std::cerr << "computing minimal open sets..." << std::endl;
    index_set_vec cs;
    h_inc.covering_sets(cs);
    index_set_vec min_open;
    for (index_type k = 0; k < cs.length(); k++) {
      hypergraph h_cand(ppcs.length());
      bool ok = true;
      for (index_type i = 0; (i < max_sat.length()) && ok; i++) {
	index_set e;
	e.fill(ppcs.length());
	e.subtract(max_sat[i]);
	e.subtract(cs[k]);
	if (e.empty()) {
	  ok = false;
	}
	else {
	  h_cand.add_edge(e);
	}
      }
      if (ok) {
	index_set_vec ccs;
	h_cand.covering_sets(ccs);
	for (index_type k = 0; k < ccs.length(); k++)
	  min_open.insert_minimal(ccs[k]);
      }
    }
    std::cerr << min_open.length() << " minimal open sets found" << std::endl;

    for (index_type k = 0; k < min_open.length(); k++) {
      if (files_written++ > file_write_limit) {
	std::cerr << "error: writing too many files!" << std::endl;
	exit(1);
      }
      std::ostringstream fname;
      char* b = reader->problem_file_basename();
      if (b) fname << b << "-";
      fname << "open" << k + 1;
      if (opt_save_compiled) fname << "-compiled";
      fname << ".pddl";
      std::cerr << "writing file " << fname.str() << "..." << std::endl;
      std::ofstream s_out(fname.str().c_str());

      s_out << ";; " << instance.name << ", open set #"
	    << (k + 1) << std::endl;
      s_out << ";; selected constraints: {";
      NTYPE val = 0;
      for (index_type i = 0; i < min_open[k].length(); i++) {
	if (i > 0) s_out << ", ";
	s_out << ppcs[min_open[k][i]].name;
	val += ppcs[min_open[k][i]].weight;
      }
      s_out << "}" << std::endl;
      s_out << ";; value = " << PRINT_NTYPE(val)
	    << ", penalty = " << PRINT_NTYPE(v_sum - val)
	    << std::endl;
      if (opt_save_compiled) {
	Instance* test_ins = instance.copy();
	for (index_type i = 0; i < min_open[k].length(); i++)
	  enforce_ppc(ppcs[min_open[k][i]], *test_ins);
	test_ins->write_domain_init(s_out);
	test_ins->write_domain_declarations(s_out);
	test_ins->write_domain_actions(s_out);
	test_ins->write_domain_DKEL_items(s_out);
	if (HSPS::Instance::write_extra) {
	  name_vec set_names(0, 0);
	  index_set_vec sets;
	  reader->export_action_partitions(set_names, sets);
	  s_out << ";; action partitions" << std::endl;
	  for (index_type k = 0; k < sets.length(); k++) {
	    instance.remap_set(sets[k], prep.action_map);
	    instance.write_domain_action_set(s_out, sets[k], set_names[k]);
	  }
	}
	s_out << ")";
	test_ins->write_problem(s_out);
	delete test_ins;
      }
      else {
	reader->write_problem_begin(s_out);
	reader->write_objects(s_out, true);
	reader->write_init(s_out);
	s_out << " (:goal (and";
	for (index_type i = 0; i < reader->dom_goals.length(); i++) {
	  if (reader->dom_goals[i]->is_state() &&
	      reader->dom_goals[i]->is_propositional()) {
	    s_out << " ";
	    reader->dom_goals[i]->print(s_out);
	  }
	}
	s_out << "))" << std::endl << " (:constraints (and";
	for (index_type i = 0; i < min_open[k].length(); i++) {
	  s_out << " ";
	  ppcs[min_open[k][i]].src->goal->print(s_out);
	}
	s_out << "))" << std::endl << ")" << std::endl;
      }
      s_out.close();
    }
  }

  if (opt_save_ext && (max_sat.length() > 0)) {
    index_set_vec exts;
    for (index_type k = 0; k < max_sat.length(); k++) {
      bool_vec in(max_sat[k], ppcs.length());
      for (index_type i = 0; i < ppcs.length(); i++) if (!in[i]) {
	in[i] = true;
	if (h_inc.is_independent(in)) {
	  exts.insert_maximal(index_set(in));
	}
	in[i] = false;
      }
    }

    std::cerr << exts.length() << " consistent 1-step extensions..."
	      << std::endl;
    for (index_type k = 0; k < exts.length(); k++) {
      if (files_written++ > file_write_limit) {
	std::cerr << "error: writing too many files!" << std::endl;
	exit(1);
      }
      std::ostringstream fname;
      char* b = reader->problem_file_basename();
      if (b) fname << b << "-";
      fname << "ext" << k + 1;
      if (opt_save_compiled) fname << "-compiled";
      fname << ".pddl";
      std::cerr << "writing file " << fname.str() << "..." << std::endl;
      std::ofstream s_out(fname.str().c_str());

      s_out << ";; " << instance.name << ", consistent extension #"
	    << (k + 1) << std::endl;
      s_out << ";; selected constraints: {";
      NTYPE val = 0;
      for (index_type i = 0; i < exts[k].length(); i++) {
	if (i > 0) s_out << ", ";
	s_out << ppcs[exts[k][i]].name;
	val += ppcs[exts[k][i]].weight;
      }
      s_out << "}" << std::endl;
      s_out << ";; value = " << PRINT_NTYPE(val)
	    << ", penalty = " << PRINT_NTYPE(v_sum - val)
	    << std::endl;

      if (opt_save_compiled) {
	Instance* test_ins = instance.copy();
	for (index_type i = 0; i < exts[k].length(); i++)
	  enforce_ppc(ppcs[exts[k][i]], *test_ins);
	test_ins->write_domain_init(s_out);
	test_ins->write_domain_declarations(s_out);
	test_ins->write_domain_actions(s_out);
	test_ins->write_domain_DKEL_items(s_out);
	if (HSPS::Instance::write_extra) {
	  name_vec set_names(0, 0);
	  index_set_vec sets;
	  reader->export_action_partitions(set_names, sets);
	  s_out << ";; action partitions" << std::endl;
	  for (index_type k = 0; k < sets.length(); k++) {
	    instance.remap_set(sets[k], prep.action_map);
	    instance.write_domain_action_set(s_out, sets[k], set_names[k]);
	  }
	}
	s_out << ")";
	test_ins->write_problem(s_out);
	delete test_ins;
      }
      else {
	reader->write_problem_begin(s_out);
	reader->write_objects(s_out, true);
	reader->write_init(s_out);
	s_out << " (:goal (and";
	for (index_type i = 0; i < reader->dom_goals.length(); i++) {
	  if (reader->dom_goals[i]->is_state() &&
	      reader->dom_goals[i]->is_propositional()) {
	    s_out << " ";
	    reader->dom_goals[i]->print(s_out);
	  }
	}
	s_out << "))" << std::endl << " (:constraints (and";
	for (index_type i = 0; i < exts[k].length(); i++) {
	  s_out << " ";
	  ppcs[exts[k][i]].src->goal->print(s_out);
	}
	s_out << "))" << std::endl << ")" << std::endl;
      }
      s_out.close();
    }
  }

  if (opt_pddl) {
    std::cout << ";; " << n_tests << " tests, " << stats << std::endl;
    std::cout << ";; max value = " << PRINT_NTYPE(v_max) << std::endl;
    std::cout << ";; min penalty = " << PRINT_NTYPE(p_min) << std::endl;

    std::cout << "(define (problem " << reader->problem_name << ")"
	      << std::endl
	      << " ;; inconsistent preference sets:" << std::endl;
    for (index_type k = 0; k < h_inc.n_edges(); k++) {
      std::cout << " (:set";
      for (index_type i = 0; i < h_inc.edge(k).length(); i++)
	std::cout << " " << ppcs[h_inc.edge(k)[i]].name;
      std::cout << ")" << std::endl;
    }
    std::cout << ")" << std::endl;
  }

  return 0;
}

END_HSPS_NAMESPACE

#ifdef USE_HSPS_NAMESPACE

int main(int argc, char *argv[])
{
  return HSPS::main(argc, argv);
}

#endif
